"""Compass Interface - Core!

This module exposes the public api for CI core
"""
from compass.core import errors
from compass.core import logger
from compass.core.hierarchy import Hierarchy
from compass.core.logon import login
from compass.core.people import People
from compass.core.reports import Reports

__all__ = (
    # sub-packages: CI-Core
    "errors",
    "logger",
    # public classes
    "Hierarchy",
    "People",
    "Reports",
    # public functions
    "login",
)



from __future__ import annotations

import re
import time
from typing import get_args, Literal, Optional, overload, TYPE_CHECKING, Union

from lxml import html

from compass.core.interface_base import InterfaceAuthenticated
from compass.core.logger import logger
from compass.core.schemas import member as schema
from compass.core.settings import Settings
from compass.core.utility import maybe_int
from compass.core.utility import parse
from compass.core.utility import validation_errors_logging

if TYPE_CHECKING:
    import datetime

    import requests

MEMBER_PROFILE_TAB_TYPES = Literal[
    "Personal", "Roles", "Permits", "Training", "Awards", "Emergency", "Comms", "Visibility", "Disclosures"
]

mogl_map = dict(
    SA="safety",
    SG="safeguarding",
    FA="first_aid",
)
mogl_types = {"gdpr", *mogl_map.values()}


class PeopleScraper(InterfaceAuthenticated):
    """Class directly interfaces with Compass operations to extract member data.

    Compass's MemberProfile.aspx has 13 tabs:
     1. Personal Details (No Key)
     2. Your Children (Page=CHILD)
     3. Roles (Page=ROLES)
     4. Permits (Page=PERMITS)
     5. Training (Page=TRAINING)
     6. Awards (Page=AWARDS)
     7. Youth Badges/Awards (Page=BADGES)
     8. Event Invitations (Page=EVENTS)
     9. Emergency Details (Page=EMERGENCY)
     10. Communications (Page=COMMS)
     11. Visibility (Page=VISIBILITY)
     12. Disclosures (Page=DISCLOSURES)
     13. Parents/Guardians (Page=PARENT)

    Of these, tabs 2, 7, 8, 13 are disabled functionality.
    Tab 11 (Visibility) is only shown on the members' own profile.

    For member-adjacent operations there are additional endpoints:
     - /Popups/Profile/AssignNewRole.aspx
     - /Popups/Maint/NewPermit.aspx
     - /Popups/Profile/EditProfile.aspx

    Currently we only use one of these endpoints (AssignNewRole), as all
    other data we need can be found from the MemberProfile tabs.

    All functions in the class output native types.
    """

    def __init__(self, session: requests.Session, member_number: int, role_number: int, jk: str):
        """Constructor for PeopleScraper.

        takes an initialised Session object from Logon
        """
        # pylint: disable=useless-super-delegation
        # Want to keep this method for future use
        super().__init__(session, member_number, role_number, jk)

    def _get_member_profile_tab(self, membership_num: int, profile_tab: MEMBER_PROFILE_TAB_TYPES) -> bytes:
        """Returns data from a given tab in MemberProfile for a given member.

        Args:
            membership_num: Membership Number to use
            profile_tab: Tab requested from Compass

        Returns:
            A dict with content and encoding, e.g.:

            {"content": b"...", "encoding": "utf-8"}

            Both keys will always be present.

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call
            ValueError: The given profile_tab value is illegal

        """
        tab_upper: str = profile_tab.upper()  # No longer type MEMBER_PROFILE_TAB_TYPES as upper case
        tabs = tuple(tab.upper() for tab in get_args(MEMBER_PROFILE_TAB_TYPES))
        url = f"{Settings.base_url}/MemberProfile.aspx?CN={membership_num}"
        if tab_upper == "PERSONAL":  # Personal tab has no key so is a special case
            response = self._get(url)
        elif tab_upper in tabs:
            url += f"&Page={tab_upper}&TAB"
            response = self._get(url)
        else:
            raise ValueError(f"Specified member profile tab {profile_tab} is invalid. Allowed values are {tabs}")

        return response.content

    def get_personal_tab(self, membership_num: int) -> schema.MemberDetails:
        """Returns data from Personal Details tab for a given member.

        Args:
            membership_num: Membership Number to use

        Returns:
            A dict mapping keys to the corresponding data from the personal
            data tab.

            E.g.:
            MemberDetails(
                membership_number=...,
                name="...",
                known_as="...",
                forenames="...",
                surname="...",
                birth_date=datetime.date(...),
                sex="...",
                nationality="...",
                ethnicity="...",
                religion="...",
                occupation="...",
                join_date=datetime.date(...),
                postcode="...",
                main_phone="...",
                main_email="..."
                address=...
            )


            Keys will be present only if valid data could be extracted and
            parsed from Compass.

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call
            PermissionError:
                Access to the member is not given by the current authentication

        """
        # pylint: disable=too-many-locals
        response = self._get_member_profile_tab(membership_num, "Personal")

        tree = html.fromstring(response)

        if tree.forms[0].action == "./ScoutsPortal.aspx?Invalid=AccessCN":
            raise PermissionError(f"You do not have permission to the details of {membership_num}")

        details: dict[str, Union[None, int, str, datetime.date]] = dict()

        # ### Extractors
        # ## Core:

        details["membership_number"] = membership_num

        # Name(s)
        names = tree.xpath("//title//text()")[0].strip().split(" ")[3:]
        details["forenames"] = names[0]
        details["surname"] = " ".join(names[1:])

        # Main Phone
        details["main_phone"] = tree.xpath('string(//*[text()="Phone"]/../../../td[3])')

        # Main Email
        details["main_email"] = tree.xpath('string(//*[text()="Email"]/../../../td[3])')

        # ## Core - Positional:

        # Full Name
        details["name"] = tree.xpath("string(//*[@id='divProfile0']//tr[1]/td[2]/label)")
        # Known As
        details["known_as"] = tree.xpath("string(//*[@id='divProfile0']//tr[2]/td[2]/label)")
        # Join Date  # TODO Unknown - take date from earliest role?
        join_date_str = tree.xpath("string(//*[@id='divProfile0']//tr[4]/td[2]/label)")
        details["join_date"] = parse(join_date_str) if join_date_str != "Unknown" else None

        # ## Position Varies, only if authorised:

        # Gender
        details["sex"] = tree.xpath("string(//*[@id='divProfile0']//*[text()='Gender:']/../../td[2])")
        # DOB
        details["birth_date"] = parse(tree.xpath("string(//*[@id='divProfile0']//*[text()='Date of Birth:']/../../td[2])"))
        # Nationality
        details["nationality"] = tree.xpath("string(//*[@id='divProfile0']//*[text()='Nationality:']/../../td[2])")
        # Ethnicity
        details["ethnicity"] = tree.xpath("normalize-space(//*[@id='divProfile0']//*[text()='Ethnicity:']/../../td[2])")
        # Religion
        details["religion"] = tree.xpath("normalize-space(//*[@id='divProfile0']//*[text()='Religion/Faith:']/../../td[2])")
        # Occupation
        details["occupation"] = tree.xpath("normalize-space(//*[@id='divProfile0']//*[text()='Occupation:']/../../td[2])")
        # Address
        original_address = tree.xpath('string(//*[text()="Address"]/../../../td[3])')
        address = original_address or ", .  "
        addr_main, addr_code = address.rsplit(". ", 1)
        postcode, country = addr_code.rsplit(" ", 1)  # Split Postcode & Country
        try:
            street, town, county = addr_main.rsplit(", ", 2)  # Split address lines
        except ValueError:
            street, town = addr_main.rsplit(", ", 1)
            county = None
        details["address"] = original_address or None
        details["country"] = country or None
        details["postcode"] = postcode or None
        details["county"] = county or None
        details["town"] = town or None
        details["street"] = street or None

        # Filter out keys with no value.
        details = {k: v for k, v in details.items() if v}
        with validation_errors_logging(membership_num):
            return schema.MemberDetails.parse_obj(details)

    def get_roles_tab(
        self, membership_num: int, keep_non_volunteer_roles: bool = False, statuses: Optional[set] = None
    ) -> schema.MemberRolesDict:
        """Returns data from Roles tab for a given member.

        Sanitises the data to a common format, and removes Occasional Helper, Network, and PVG roles by default.

        Args:
            membership_num: Membership Number to use
            keep_non_volunteer_roles: Keep Helper (OH/PVG) & Network roles?
            statuses: Explicit set of role statuses to keep

        Returns:
            A dict of dicts mapping keys to the corresponding data from the roles tab.

            E.g.:
            {1234578:
             {'role_number': 1234578,
              'membership_number': ...,
              'role_title': '...',
              'role_class': '...',
              'role_type': '...',
              'location_id': ...,
              'location_name': '...',
              'role_start_date': datetime.datetime(...),
              'role_end': datetime.datetime(...),
              'role_status': '...'},
             {...}
            }


            Keys will always be present.

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call
            PermissionError:
                Access to the member is not given by the current authentication

        Todo:
            primary_role

        """
        # pylint: disable=too-many-locals
        # Want to keep all functionality in one place, to reduce the number of
        # calls to Compass.
        # TODO could refactor some internals into helper functions
        logger.debug(f"getting roles tab for member number: {membership_num}")
        response = self._get_member_profile_tab(membership_num, "Roles")
        tree = html.fromstring(response)

        if tree.forms[0].action == "./ScoutsPortal.aspx?Invalid=AccessCN":
            raise PermissionError(f"You do not have permission to the details of {membership_num}")

        statuses_set = statuses is not None

        roles_data = {}
        rows = tree.xpath("//tbody/tr")
        for row in rows:
            # Get children (cells in row)
            cells = list(row)  # filter out empty elements

            # If current role allows selection of role for editing, remove tickbox
            if any(el.tag == "input" for el in cells[0]):
                cells.pop(0)

            role_number = int(row.get("data-pk"))
            status_with_review = cells[5].text_content().strip()
            if status_with_review.startswith("Full Review Due "):
                role_status = "Full"
                review_date = parse(status_with_review.removeprefix("Full Review Due "))
            else:
                role_status = status_with_review
                review_date = None

            role_details = dict(
                role_number=role_number,
                membership_number=membership_num,
                role_title=cells[0].text_content().strip(),
                role_class=cells[1].text_content().strip(),
                # role_type only visible if access to System Admin tab
                role_type=[*row.xpath("./td[1]/*/@title"), None][0],
                # location_id only visible if role is in hierarchy AND location still exists
                location_id=cells[2][0].get("data-ng_id"),
                location_name=cells[2].text_content().strip(),
                role_start=parse(cells[3].text_content().strip()),
                role_end=parse(cells[4].text_content().strip()),
                role_status=role_status,
                review_date=review_date,
                can_view_details=any("VIEWROLE" in el.get("class") for el in cells[6]),
            )
            # Remove OHs etc from list
            if not keep_non_volunteer_roles and (
                "helper" in role_details["role_class"].lower()
                or {role_details["role_title"].lower()} <= {"occasional helper", "pvg", "network member"}
            ):
                continue

            # Role status filter
            if statuses_set and role_status not in statuses:
                continue

            roles_data[role_number] = role_details

        with validation_errors_logging(membership_num):
            return schema.MemberRolesDict.parse_obj(roles_data)

    def get_permits_tab(self, membership_num: int) -> schema.MemberPermitsList:
        """Returns data from Permits tab for a given member.

        If a permit has been revoked, the expires value is None and the status is PERM_REV

        Args:
            membership_num: Membership Number to use

        Returns:
            A list of dicts mapping keys to the corresponding data from the
            permits tab.

            Keys will always be present.

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call

        """
        response = self._get_member_profile_tab(membership_num, "Permits")
        tree = html.fromstring(response)

        # Get rows with permit content
        rows = tree.xpath('//table[@id="tbl_p4_permits"]//tr[@class="msTR msTRPERM"]')

        permits = []
        for row in rows:
            permit: dict[str, Union[None, int, str, datetime.date]] = dict(membership_number=membership_num)
            child_nodes = list(row)
            permit["permit_type"] = child_nodes[1].text_content()
            permit["category"] = child_nodes[2].text_content()
            permit["type"] = child_nodes[3].text_content()
            permit["restrictions"] = child_nodes[4].text_content()
            expires = child_nodes[5].text_content()
            permit["expires"] = parse(expires) if expires != "Revoked" else None
            permit["status"] = child_nodes[5].get("class")

            permits.append(permit)

        with validation_errors_logging(membership_num):
            return schema.MemberPermitsList.parse_obj(permits)

    @overload
    def get_training_tab(self, membership_num: int, ongoing_only: Literal[True]) -> schema.MemberMOGLList:
        ...

    @overload
    def get_training_tab(self, membership_num: int, ongoing_only: Literal[False]) -> schema.MemberTrainingTab:
        ...

    @overload
    def get_training_tab(self, membership_num: int, ongoing_only: bool) -> Union[schema.MemberTrainingTab, schema.MemberMOGLList]:
        ...

    def get_training_tab(
        self, membership_num: int, ongoing_only: bool = False
    ) -> Union[schema.MemberTrainingTab, schema.MemberMOGLList]:
        """Returns data from Training tab for a given member.

        Args:
            membership_num: Membership Number to use
            ongoing_only: Return a dataframe of role training & OGL info? Otherwise returns all data

        Returns:
            A dict mapping keys to the corresponding data from the training
            tab.

            E.g.:
            {'roles': {1234567: {'role_number': 1234567,
               'role_title': '...',
               'role_start': datetime.datetime(...),
               'role_status': '...',
               'location': '...',
               'ta_data': '...',
               'ta_number': '...',
               'ta_name': '...',
               'completion': '...',
               'wood_badge_number': '...'},
              ...},
             'plps': {1234567: [{'pk': 6142511,
                'module_id': ...,
                'code': '...',
                'name': '...',
                'learning_required': False,
                'learning_method': '...',
                'learning_completed': '...',
                'validated_membership_number': '...',
                'validated_name': '...'},
               ...],
              ...},
             'mandatory': {'GDPR':
              {'name': 'GDPR',
              'completed_date': datetime.datetime(...)},
              ...}}

            Keys will always be present.

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call

        """
        # pylint: disable=too-many-locals,too-many-statements
        # Want to keep all functionality in one place, to reduce the number of
        # calls to Compass.
        # TODO could refactor some internals into helper functions
        logger.debug(f"getting training tab for member number: {membership_num}")

        response = self._get_member_profile_tab(membership_num, "Training")
        tree = html.fromstring(response)

        rows = tree.xpath("//table[@id='tbl_p5_TrainModules']/tr")

        training_plps = {}
        training_roles = {}
        for row in rows:
            # Personal Learning Plan (PLP) data
            if "trPLP" in row.classes:
                plp = row
                plp_table = plp.getchildren()[0].getchildren()[0]
                plp_data = []
                for module_row in plp_table:
                    if module_row.get("class") != "msTR trMTMN":
                        continue

                    module_data: dict[str, Union[None, int, str, datetime.date]] = {}
                    child_nodes = list(module_row)
                    module_data["pk"] = int(module_row.get("data-pk"))
                    module_data["module_id"] = int(child_nodes[0].get("id")[4:])
                    matches = re.match(r"^([A-Z0-9]+) - (.+)$", child_nodes[0].text_content()).groups()
                    if matches:
                        code = str(matches[0])
                        module_data["code"] = code
                        module_data["name"] = matches[1]

                        # Skip processing if we only want ongoing learning data and the module is not GDPR.
                        if ongoing_only and "gdpr" not in code.lower():
                            continue

                    learning_required = child_nodes[1].text_content().lower()
                    module_data["learning_required"] = "yes" in learning_required if learning_required else None
                    module_data["learning_method"] = child_nodes[2].text_content() or None
                    module_data["learning_completed"] = parse(child_nodes[3].text_content())
                    module_data["learning_date"] = parse(child_nodes[3].text_content())

                    validated_by_string = child_nodes[4].text_content()
                    if validated_by_string:
                        # Add empty item to prevent IndexError
                        validated_by_data = validated_by_string.split(" ", maxsplit=1) + [""]
                        module_data["validated_membership_number"] = maybe_int(validated_by_data[0])
                        module_data["validated_name"] = validated_by_data[1]
                    module_data["validated_date"] = parse(child_nodes[5].text_content())

                    plp_data.append(module_data)

                training_plps[int(plp_table.get("data-pk"))] = plp_data

            # Role data
            if "msTR" in row.classes:
                role = row

                child_nodes = list(role)

                info: dict[str, Union[None, str, int, datetime.date]] = {}  # NoQA

                info["role_number"] = int(role.xpath("./@data-ng_mrn")[0])
                info["role_title"] = child_nodes[0].text_content()
                info["role_start"] = parse(child_nodes[1].text_content())
                status_with_review = child_nodes[2].text_content()
                # TODO for `Ending: blah` roles, should we store the ending date?
                if status_with_review.startswith("Full (Review Due: ") or status_with_review.startswith("Full (Ending: "):
                    info["role_status"] = "Full"
                    date_str = status_with_review.removeprefix("Full (Review Due: ").removeprefix("Full (Ending: ").rstrip(")")
                    info["review_date"] = parse(date_str)
                else:
                    info["role_status"] = status_with_review
                    info["review_date"] = None

                info["location"] = child_nodes[3].text_content()

                training_advisor_string = child_nodes[4].text_content()
                if training_advisor_string:
                    info["ta_data"] = training_advisor_string
                    # Add empty item to prevent IndexError
                    training_advisor_data = training_advisor_string.split(" ", maxsplit=1) + [""]
                    info["ta_number"] = maybe_int(training_advisor_data[0])
                    info["ta_name"] = training_advisor_data[1]

                completion_string = child_nodes[5].text_content()
                if completion_string:
                    info["completion"] = completion_string
                    parts = completion_string.split(":")
                    info["completion_type"] = parts[0].strip()
                    info["completion_date"] = parse(parts[1].strip())
                    assert len(parts) <= 2, parts[2:]
                    # info["ct"] = parts[3:]  # TODO what is this? From CompassRead.php
                info["wood_badge_number"] = child_nodes[5].get("id", "").removeprefix("WB_") or None

                training_roles[info["role_number"]] = info

        # Handle GDPR:
        # Get latest GDPR date
        gdpr_dates = [mod["validated_date"] for plp in training_plps.values() for mod in plp if mod["code"] == "GDPR"]
        training_ogl = {"gdpr": dict(completed_date=next(reversed(sorted(date for date in gdpr_dates if date is not None)), None))}
        for ongoing_learning in tree.xpath("//tr[@data-ng_code]"):
            cell_text = {c.get("id", "<None>").split("_")[0]: c.text_content() for c in ongoing_learning}

            training_ogl[mogl_map[ongoing_learning.get("data-ng_code")]] = dict(
                completed_date=parse(cell_text.get("tdLastComplete")),
                renewal_date=parse(cell_text.get("tdRenewal")),
            )
            # TODO missing data-pk from list(cell)[0].tag == "input", and module names/codes. Are these important?

        # Update training_ogl with missing mandatory ongoing learning types
        training_ogl |= {missing_mogl_type: dict() for missing_mogl_type in mogl_types - training_ogl.keys()}

        if ongoing_only:
            with validation_errors_logging(membership_num):
                return schema.MemberMOGLList.parse_obj(training_ogl)

        training_data = {
            "roles": training_roles,
            "plps": training_plps,
            "mandatory": training_ogl,
        }

        with validation_errors_logging(membership_num):
            return schema.MemberTrainingTab.parse_obj(training_data)

    def get_awards_tab(self, membership_num: int) -> list[schema.MemberAward]:
        """Returns data from Awards tab for a given member.

        Args:
            membership_num: Membership Number to use

        Returns:
            A MemberAward object with corresponding data from the awards tab.

            E.g.:
            MemberAward(
                membership_number=...,
                type="...",
                location="...",
                date=datetime.date(...),
            )

            Keys will always be present.

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call

        """
        response = self._get_member_profile_tab(membership_num, "Awards")
        tree = html.fromstring(response)

        if tree.forms[0].action == "./ScoutsPortal.aspx?Invalid=AccessCN":
            raise PermissionError(f"You do not have permission to the details of {membership_num}")

        awards = []
        rows = tree.xpath("//table[@class='msAward']/tr")
        with validation_errors_logging(membership_num):
            for row in rows:
                award_props = row[1][0]  # Properties are stored as yet another sub-table
                award_data = schema.MemberAward(
                    membership_number=membership_num,
                    type=award_props[0][1].text_content(),
                    location=award_props[1][1].text_content() or None,
                    date=parse(award_props[2][1].text_content() or ""),
                )
                awards.append(award_data)
        return awards

    def get_disclosures_tab(self, membership_num: int) -> list[schema.MemberDisclosure]:
        """Returns data from Disclosures tab for a given member.

        Args:
            membership_num: Membership Number to use

        Returns:
            A MemberAward object with corresponding data from the disclosures
            tab.

            E.g.:
            MemberDisclosure(
                membership_number=...,
                country="...",
                provider="...",
                type="...",
                number=...,
                issuer="...",
                issue_date=datetime.date(...),
                status="...",
                expiry_date=datetime.date(...),
            )

            Keys will always be present.

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call

        """
        response = self._get_member_profile_tab(membership_num, "Disclosures")
        tree = html.fromstring(response)

        if tree.forms[0].action == "./ScoutsPortal.aspx?Invalid=AccessCN":
            raise PermissionError(f"You do not have permission to the details of {membership_num}")

        disclosures = []
        rows = tree.xpath("//tbody/tr")
        with validation_errors_logging(membership_num):
            for row in rows:
                # Get children (cells in row)
                cells = list(row)

                disclosure = schema.MemberDisclosure(
                    membership_number=membership_num,
                    country=cells[0].text_content() or None,  # Country sometimes missing (Application Withdrawn)
                    provider=cells[1].text_content(),
                    type=cells[2].text_content(),
                    number=cells[3].text_content() or None,  # If Application Withdrawn, no disclosure number
                    issuer=cells[4].text_content() or None,
                    issue_date=parse(cells[5].text_content()),  # If Application Withdrawn, maybe no issue date
                    status=cells[6].text_content(),
                    expiry_date=parse(cells[7].text_content()),  # If Application Withdrawn, no expiry date
                )
                disclosures.append(disclosure)
        return disclosures

    # See getAppointment in PGS\Needle
    def get_roles_detail(
        self, role_number: int, response: Union[None, str, bytes, requests.Response] = None
    ) -> schema.MemberRolePopup:
        """Returns detailed data from a given role number.

        Args:
            role_number: Role Number to use
            response: Pre-generated response to use

        Returns:
            A dicts mapping keys to the corresponding data from the
            role detail data.

            E.g.:
            {'hierarchy': {'organisation': 'The Scout Association',
              'country': '...',
              'region': '...',
              'county': '...',
              'district': '...',
              'group': '...',
              'section': '...'},
             'details': {'role_number': ...,
              'organisation_level': '...',
              'birth_date': datetime.datetime(...),
              'membership_number': ...,
              'name': '...',
              'role_title': '...',
              'role_start': datetime.datetime(...),
              'role_status': '...',
              'line_manager_number': ...,
              'line_manager': '...',
              'ce_check': datetime.datetime(...),
              'disclosure_check': '...',
              'references': '...',
              'appointment_panel_approval': '...',
              'commissioner_approval': '...',
              'committee_approval': '...'},
             'getting_started': {...: {'name': '...',
               'validated': datetime.datetime(...),
               'validated_by': '...'},
               ...
              }}

            Keys will always be present.

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call

        """
        # pylint: disable=too-many-locals,too-many-statements
        # Want to keep all functionality in one place, to reduce the number of
        # calls to Compass.
        # TODO could refactor some internals into helper functions
        renamed_levels = {
            "County / Area / Scottish Region / Overseas Branch": "County",
        }
        renamed_modules = {
            "001": "module_01",
            "TRST": "trustee_intro",
            "002": "module_02",
            "003": "module_03",
            "004": "module_04",
            "GDPR": "GDPR",
            "SFTY": "safety",
            "SAFE": "safeguarding",
        }
        unset_vals = {"--- Not Selected ---", "--- No Items Available ---", "--- No Line Manager ---"}

        module_names = {
            "Essential Information": "M01",
            "Trustee Introduction": "TRST",
            "Personal Learning Plan": "M02",
            "Tools for the Role (Section Leaders)": "M03",
            "Tools for the Role (Managers and Supporters)": "M04",
            "General Data Protection Regulations": "GDPR",
            "Safety Training": "SFTY",
            "Safeguarding Training": "SAFE",
        }

        references_codes = {
            "NC": "Not Complete",
            "NR": "Not Required",
            "RR": "References Requested",
            "S": "References Satisfactory",
            "U": "References Unsatisfactory",
        }

        start_time = time.time()
        if response is None:
            response = self._get(f"{Settings.base_url}/Popups/Profile/AssignNewRole.aspx?VIEW={role_number}")
            logger.debug(f"Getting details for role number: {role_number}. Request in {(time.time() - start_time):.2f}s")

        post_response_time = time.time()
        if isinstance(response, (str, bytes)):
            tree = html.fromstring(response)
        else:
            tree = html.fromstring(response.content)
        form = tree.forms[0]
        inputs = form.inputs
        fields = form.fields

        if form.action == "./ScoutsPortal.aspx?Invalid=Access":
            raise PermissionError(f"You do not have permission to the details of role {role_number}")

        member_string = fields.get("ctl00$workarea$txt_p1_membername")
        ref_code = fields.get("ctl00$workarea$cbo_p2_referee_status")

        role_details: dict[str, Union[None, int, str, datetime.date]] = dict()
        # Approval and Role details
        role_details["role_number"] = role_number
        role_details["organisation_level"] = fields.get("ctl00$workarea$cbo_p1_level")
        role_details["birth_date"] = parse(inputs["ctl00$workarea$txt_p1_membername"].get("data-dob")) if Settings.debug else None
        role_details["membership_number"] = int(fields.get("ctl00$workarea$txt_p1_memberno"))
        role_details["name"] = member_string.split(" ", maxsplit=1)[1]  # TODO does this make sense - should name be in every role??
        role_details["role_title"] = fields.get("ctl00$workarea$txt_p1_alt_title")
        role_details["role_start"] = parse(fields.get("ctl00$workarea$txt_p1_startdate"))
        # Role Status
        role_details["role_status"] = fields.get("ctl00$workarea$txt_p2_status")
        # Line Manager
        line_manager_el = next((op for op in inputs["ctl00$workarea$cbo_p2_linemaneger"] if op.get("selected")), None)
        role_details["line_manager_number"] = maybe_int(line_manager_el.get("value")) if line_manager_el is not None else None
        role_details["line_manager"] = line_manager_el.text.strip() if line_manager_el is not None else None
        # Review Date
        role_details["review_date"] = parse(fields.get("ctl00$workarea$txt_p2_review"))
        # CE (Confidential Enquiry) Check  # TODO if CE check date != current date then is valid
        ce_check = fields.get("ctl00$workarea$txt_p2_cecheck")
        role_details["ce_check"] = parse(ce_check) if ce_check != "Pending" else None
        # Disclosure Check
        disclosure_with_date = fields.get("ctl00$workarea$txt_p2_disclosure", "")
        if disclosure_with_date.startswith("Disclosure Issued : "):
            disclosure_date = parse(disclosure_with_date.removeprefix("Disclosure Issued : "))
            disclosure_check = "Disclosure Issued"
        else:
            disclosure_date = None
            disclosure_check = disclosure_with_date or None
        role_details["disclosure_check"] = disclosure_check
        role_details["disclosure_date"] = disclosure_date
        # References
        role_details["references"] = references_codes.get(ref_code, ref_code)

        approval_values = {}
        for row in tree.xpath("//tr[@class='trProp']"):
            select = row[1][0]
            code = select.get("data-app_code")
            approval_values[code] = select.get("data-db")
            # select.get("title") gives title text, but this is not useful as it does not reflect latest changes,
            # but only who added the role to Compass.

        # Appointment Panel Approval
        role_details["appointment_panel_approval"] = approval_values.get("ROLPRP|AACA")
        # Commissioner Approval
        role_details["commissioner_approval"] = approval_values.get("ROLPRP|CAPR")
        # Committee Approval
        role_details["committee_approval"] = approval_values.get("ROLPRP|CCA")

        if role_details["line_manager_number"] in unset_vals:
            role_details["line_manager_number"] = None

        # Filter null values
        role_details = {k: v for k, v in role_details.items() if v is not None}

        # Getting Started
        modules_output = {}
        getting_started_modules = tree.xpath("//tr[@class='trTrain trTrainData']")
        # Get all training modules and then extract the required modules to a dictionary
        for module in getting_started_modules:
            module_name = module[0][0].text.strip()
            if module_name in module_names:
                info = {
                    # "name": module_names[module_name],  # short_name
                    "validated": parse(module[2][0].value),  # Save module validation date
                    "validated_by": module[1][1].get("value") or None,  # Save who validated the module
                }
                mod_code: str = module[2][0].get("data-ng_value")
                modules_output[renamed_modules[mod_code]] = info

        # Get all levels of the org hierarchy and select those that will have information:
        # Get all inputs with location data
        org_levels = [v for k, v in sorted(dict(inputs).items()) if "ctl00$workarea$cbo_p1_location" in k]
        # TODO
        all_locations = {row.get("title"): row.findtext("./option") for row in org_levels}

        clipped_locations = {
            renamed_levels.get(key, key).lower(): value for key, value in all_locations.items() if value not in unset_vals
        }

        logger.debug(
            f"Processed details for role number: {role_number}. "
            f"Compass: {(post_response_time - start_time):.3f}s; Processing: {(time.time() - post_response_time):.4f}s"
        )
        # TODO data-ng_id?, data-rtrn_id?
        full_details = {
            "hierarchy": clipped_locations,
            "details": role_details,
            "getting_started": modules_output,
        }
        with validation_errors_logging(role_number, name="Role Number"):
            return schema.MemberRolePopup.parse_obj(full_details)



from __future__ import annotations

import contextlib
import enum
import json
from pathlib import Path
from typing import Iterable, Literal, Optional, TYPE_CHECKING, Union

from pydantic.json import pydantic_encoder

from compass.core import utility
from compass.core._scrapers.hierarchy import HierarchyScraper
from compass.core.logger import logger
from compass.core.logon import Logon
from compass.core.schemas import hierarchy as schema

if TYPE_CHECKING:
    from collections.abc import Iterator

TYPES_UNIT_LEVELS = Literal["Group", "District", "County", "Region", "Country", "Organisation"]


class Levels(enum.IntEnum):
    Organisation = 1
    Country = 2
    Region = 3
    County = 4
    District = 5
    Group = 6


class UnitChildren(enum.IntEnum):
    countries = Levels.Organisation
    regions = Levels.Country
    counties = Levels.Region
    districts = Levels.County
    groups = Levels.District


class UnitSections(enum.IntEnum):
    hq_sections = Levels.Organisation
    country_sections = Levels.Country
    region_sections = Levels.Region
    county_sections = Levels.County
    district_sections = Levels.District
    group_sections = Levels.Group


class Hierarchy:
    def __init__(self, session: Logon):
        """Constructor for Hierarchy."""
        self._scraper: HierarchyScraper = HierarchyScraper(session.s, session.cn, session.mrn, session.jk)
        self.session: Logon = session

    def get_unit_data(
        self,
        unit_level: Optional[schema.HierarchyLevel] = None,
        _id: Optional[int] = None,
        level: Optional[TYPES_UNIT_LEVELS] = None,
        use_default: bool = False,
    ) -> schema.HierarchyLevel:
        """Helper function to construct unit level data.

        Unit data can be specified as a pre-constructed model, by passing literals, or
        by signalling to use the data from the user's current role. If all three
        options are unset an exception is raised.

        There is a strict priority ordering as follows:
            1. pre-constructed pydantic model
            2. literals
            3. default data

        Returns:
            Constructed unit level data, as a pydantic model.
            e.g.:
                HierarchyLevel(id=..., level="...")

        Raises:
            ValueError:
                When no unit data information has been provided

        """
        if unit_level is not None:
            data = unit_level
        elif _id is not None and level is not None:
            data = schema.HierarchyLevel(id=_id, level=level)
        elif use_default:
            data = self.session.hierarchy  # as this is a property, it will update when roles change
        else:
            raise ValueError("No level data specified! unit_level, id and level, or use_default must be set!")

        logger.debug(f"found unit data: id: {data.id}, level: {data.level}")

        return data

    # See recurseRetrieve in PGS\Needle
    def get_hierarchy(
        self,
        unit_level: Optional[schema.HierarchyLevel] = None,
        unit_id: Optional[int] = None,
        level: Optional[TYPES_UNIT_LEVELS] = None,
        use_default: bool = False,
    ) -> schema.UnitData:
        """Gets all units at given level and below, including sections.

        Unit data can be specified as a pre-constructed model, by passing literals, or
        by signalling to use the data from the user's current role. If all three
        options are unset an exception is raised.

        There is a strict priority ordering as follows:
            1. pre-constructed pydantic model
            2. literals
            3. default data

        Raises:
            ValueError:
                When no unit data information has been provided

        """
        unit_level = self.get_unit_data(unit_level, unit_id, level, use_default)

        filename = Path(f"hierarchy-{unit_level.id}.json")
        # Attempt to see if the hierarchy has been fetched already and is on the local system
        with contextlib.suppress(FileNotFoundError):
            out = json.loads(filename.read_text(encoding="utf-8"))
            if isinstance(out, dict):
                return schema.UnitData.parse_obj(out)

        # Fetch the hierarchy
        out = self._get_descendants_recursive(unit_level.id, hier_level=unit_level.level)

        # Try and write to a file for caching
        with utility.filesystem_guard("Unable to write cache file"):
            filename.write_text(schema.UnitData.parse_obj(out).json(ensure_ascii=False), encoding="utf-8")

        return schema.UnitData.parse_obj(out)

    # See recurseRetrieve in PGS\Needle
    def _get_descendants_recursive(
        self, compass_id: int, hier_level: Optional[TYPES_UNIT_LEVELS] = None, hier_num: Optional[Levels] = None
    ) -> dict[str, object]:
        """Recursively get all children from given unit ID and level name/number, with caching."""
        if hier_num is not None:
            level_numeric = hier_num
        elif hier_level is not None:
            try:
                level_numeric = Levels[hier_level]
            except KeyError:
                valid_levels = [level.name for level in Levels]
                raise ValueError(f"Passed level: {hier_level} is illegal. Valid values are {valid_levels}")
        else:
            raise ValueError("A numeric or string hierarchy level needs to be passed")

        logger.debug(f"getting data for unit {compass_id}")
        descendants = level_numeric in set(UnitChildren)  # Do child units exist? (i.e. is this level != group)

        # All to handle as Group doesn't have grand-children
        descendant_data = {"id": compass_id, "level": level_numeric.name}

        child_level = Levels(level_numeric + 1) if descendants else None
        if descendants:
            children = self._scraper.get_units_from_hierarchy(compass_id, UnitChildren(level_numeric).name)
            children_updated = []
            for child in children:
                if not child:
                    continue
                grandchildren = self._get_descendants_recursive(child.id, hier_num=child_level)
                children_updated.append(child.dict() | grandchildren)
            descendant_data["child"] = children_updated
        descendant_data["sections"] = self._scraper.get_units_from_hierarchy(compass_id, UnitSections(level_numeric).name)

        return descendant_data

    @staticmethod
    def flatten_hierarchy(hierarchy_dict: schema.UnitData) -> Iterator[dict[str, Union[int, str]]]:  # noqa: D417
        """Flattens a hierarchy tree / graph to a flat sequence of mappings.

        Args:
            hierarchy_dict:
                The current object to be flattened. The user will pass in a `schema.UnitData`
                object, whilst all recursion will be on `schema.DescendantData` objects.

        """
        # This args style is allowed, but not yet (2021-02-20) implemented in PyDocStyle, so D417 disabled above.
        # https://github.com/PyCQA/pydocstyle/issues/449

        def flatten(
            d: Union[schema.UnitData, schema.DescendantData], hierarchy_state: dict[str, Union[int, str]]
        ) -> Iterator[dict[str, Union[int, str]]]:
            """Generator expresion to recursively flatten hierarchy."""
            level_name = d.level
            compass_id = d.id
            name = d.name if "name" in hierarchy_dict.__fields__ else None  # ~900ns
            level_data = {
                **hierarchy_state,
                f"{level_name}_ID": compass_id,
                f"{level_name}_name": name,
            }
            yield {"compass": compass_id, "name": name, **level_data}
            for val in d.child or []:
                yield from flatten(val, level_data)
            for val in d.sections:
                yield {"compass": val.id, "name": val.name, **level_data}  # TODO section=True flag?

        return flatten(hierarchy_dict, dict())

    def get_unique_members(
        self,
        unit_level: Optional[schema.HierarchyLevel] = None,
        unit_id: Optional[int] = None,
        level: Optional[TYPES_UNIT_LEVELS] = None,
        use_default: bool = False,
    ) -> set[int]:
        """Get all unique members for a given level and its descendants.

        Unit data can be specified as a pre-constructed model, by passing literals, or
        by signalling to use the data from the user's current role. If all three
        options are unset an exception is raised.

        There is a strict priority ordering as follows:
            1. pre-constructed pydantic model
            2. literals
            3. default data

        Returns:
            A set of unique member numbers within the given unit.

        Raises:
            ValueError:
                When no unit data information has been provided

        """
        unit_level = self.get_unit_data(unit_level, unit_id, level, use_default)

        # get tree of all units
        hierarchy_dict = self.get_hierarchy(unit_level)

        # flatten tree
        flat_hierarchy = self.flatten_hierarchy(hierarchy_dict)

        # generator for compass unit IDs
        compass_ids = (unit["compass"] for unit in flat_hierarchy)

        # get members from the list of IDs
        units_members = self.get_members_in_units(unit_level.id, compass_ids)

        # return a set of membership numbers
        return {members.contact_number for unit_members in units_members for members in unit_members.member}

    def get_members_in_units(self, parent_id: int, compass_ids: Iterable[int]) -> list[schema.HierarchyUnitMembers]:
        filename = Path(f"all-members-{parent_id}.json")

        with contextlib.suppress(FileNotFoundError):
            # Attempt to see if the members dict has been fetched already and is on the local system
            all_members = json.loads(filename.read_text(encoding="utf-8"))
            if isinstance(all_members, list):
                return [schema.HierarchyUnitMembers(**unit_members) for unit_members in all_members]

        # Fetch all members
        all_members = []
        for compass_id in set(compass_ids):
            logger.debug(f"Getting members for {compass_id}")
            all_members.append(dict(compass_id=compass_id, member=self._scraper.get_members_with_roles_in_unit(compass_id)))

        # Try and write to a file for caching
        with utility.filesystem_guard("Unable to write cache file"):
            filename.write_text(json.dumps(all_members, ensure_ascii=False, indent=4, default=pydantic_encoder), encoding="utf-8")

        return schema.HierarchyUnitMembersList.parse_obj(all_members).__root__



from __future__ import annotations

from typing import Literal, Optional

import pydantic

TYPES_SECTION = Literal[
    "Early Years Pilot",
    "Beavers",
    "Beaver Scout",
    "Cub Scout",
    "Explorer Scouts",
    "Scout",
    "Scout Active Support",
    "Scout Network",
    "Other",
]
TYPES_UNIT_LEVELS = Literal["Group", "District", "County", "Region", "Country", "Organisation"]


class HierarchyBase(pydantic.BaseModel):
    id: int


class HierarchyUnit(HierarchyBase):
    name: str
    parent_id: int

    # metadata
    status: Literal["ACT"]
    address: str
    member_count: int


class HierarchySection(HierarchyUnit):
    # section details (only if a section!)
    section_type: Optional[TYPES_SECTION]


class HierarchyLevel(HierarchyBase):
    level: TYPES_UNIT_LEVELS


class UnitData(HierarchyLevel):
    child: Optional[list[DescendantData]]  # NOTE: deliberate recursive/forward reference here!
    sections: list[HierarchySection]


class DescendantData(HierarchyUnit, UnitData):
    pass


UnitData.update_forward_refs()  # NOTE: updating recursive/forward reference here!
DescendantData.update_forward_refs()  # NOTE: updating recursive/forward reference here!


class HierarchyMember(pydantic.BaseModel):
    contact_number: int
    name: Optional[str]
    role: Optional[str]


class HierarchyUnitMembers(pydantic.BaseModel):
    compass_id: int  # TODO disambiguate this
    member: list[HierarchyMember]


class HierarchyUnitMembersList(pydantic.BaseModel):
    __root__: list[HierarchyUnitMembers]



from __future__ import annotations

from typing import get_args, Optional, TYPE_CHECKING

from compass.core._scrapers.member import PeopleScraper
from compass.core.schemas.member import TYPES_ROLE_STATUS

if TYPE_CHECKING:
    from collections.abc import Iterable

    from compass.core.logon import Logon
    from compass.core.schemas import member as schema

# SCRAPER CLASS - 1-1 mapping with compass to minimise calls
# MAIN CLASS - object/properties focused, with abstractions of actual calls
# UTILITY CLASS - get_member_data, get_roles_from_members, etc

STATUSES = set(get_args(TYPES_ROLE_STATUS))


class People:
    def __init__(self, session: Logon):
        """Constructor for People."""
        self._scraper = PeopleScraper(session.s, session.cn, session.mrn, session.jk)

    def personal(self, membership_num: int) -> schema.MemberDetails:
        """Gets personal tab data for a given member.

        Args:
            membership_num: Membership Number to use

        Returns:
            A MemberDetails object containing all data.

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call
            PermissionError:
                If the current user does not have permission to view personal
                data for the requested member.

        """
        return self._scraper.get_personal_tab(membership_num)

    # See getRole in PGS\Needle
    def roles(
        self,
        membership_num: int,
        keep_non_volunteer_roles: bool = False,
        only_active: bool = False,
        statuses: Optional[Iterable[TYPES_ROLE_STATUS]] = None,
    ) -> schema.MemberRolesDict:
        """Gets the data from the Role tab in Compass for the specified member.

        Sanitises the data to a common format, and removes Occasional Helper, Network, and PVG roles by default.

        Args:
            membership_num: Membership Number to use
            keep_non_volunteer_roles: Keep Helper (OH/PVG) & Network roles?
            only_active: Keep only active (Full, Provisional, Pre-Provisional) roles?
            statuses: Explicit set of role statuses to keep

        Returns:
            A MemberRolesDict object containing all data.

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call
            PermissionError:
                If the current user does not have permission to view roles
                data for the requested member.

        """
        if only_active:
            statuses = STATUSES - {"Closed", "Cancelled"}
        return self._scraper.get_roles_tab(membership_num, keep_non_volunteer_roles, statuses)

    def role_detail(self, role_number: int) -> schema.MemberRolePopup:
        """Get detailed information for specified role.

         Make sure to check `MemberRoleCore.can_view_details` on each role, as
         you may not have permissions to view certain roles (if they are
         outside of your hierarchy, for example).

         Args:
             role_number: Role to query

        Returns:
            MemberRolePopup object with detail on the role

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call
            PermissionError:
                If the current user does not have permission to view role data

        """
        return self._scraper.get_roles_detail(role_number)

    def permits(self, membership_num: int) -> schema.MemberPermitsList:
        """Gets permits tab data for a given member.

        Args:
            membership_num: Membership Number to use

        Returns:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call
            A MemberPermitsList object containing all data.

        """
        return self._scraper.get_permits_tab(membership_num)

    def training(self, membership_num: int) -> schema.MemberTrainingTab:
        """Gets training tab data for a given member.

        Args:
            membership_num: Membership Number to use

        Returns:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call
            A MemberTrainingTab object containing all data.

        """
        return self._scraper.get_training_tab(membership_num, ongoing_only=False)

    def ongoing_learning(self, membership_num: int) -> schema.MemberMOGLList:
        """Gets ongoing learning data for a given member.

        Args:
            membership_num: Membership Number to use

        Returns:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call
            A MemberMOGLList object containing all data.

        """
        return self._scraper.get_training_tab(membership_num, ongoing_only=True)

    def awards(self, membership_num: int) -> list[schema.MemberAward]:
        """Gets awards tab data for a given member.

        Args:
            membership_num: Membership Number to use

        Returns:
            A MemberAward object containing all data.

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call
            PermissionError:
                If the current user does not have permission to view awards
                data for the requested member.

        """
        return self._scraper.get_awards_tab(membership_num)

    def disclosures(self, membership_num: int) -> list[schema.MemberDisclosure]:
        """Gets disclosures tab data for a given member.

        Args:
            membership_num: Membership Number to use

        Returns:
            A list of MemberDisclosure objects containing all data.

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call
            PermissionError:
                If the current user does not have permission to view
                disclosure data for the requested member.

        """
        return self._scraper.get_disclosures_tab(membership_num)

    def latest_disclosure(self, membership_num: int) -> Optional[schema.MemberDisclosure]:
        """Gets latest disclosure for a given member.

        Latest is defined as the expiry date that is furthest away.

        Args:
            membership_num: Membership Number to use

        Returns:
            A MemberDisclosure objects containing disclosure data.

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call
            PermissionError:
                If the current user does not have permission to view
                disclosure data for the requested member.

        """
        disclosures = self.disclosures(membership_num)
        date_map = {disc.expiry_date: disc for disc in disclosures if disc.expiry_date}
        return date_map.get(max(date_map.keys(), default=None))


# class Member:
#     def personal_details(self):
#         pass
#
#     def role_details(self):
#         pass
#
#     def ongoing_learning(self):
#         pass
#
#     def training(self):
#         pass
#
#     def permits(self):
#         pass
#
#     # def awards(self):
#     #     pass



from __future__ import annotations

import json
import typing
from typing import Literal, TYPE_CHECKING, Union

from lxml import html
import pydantic

from compass.core.errors import CompassError
from compass.core.interface_base import InterfaceAuthenticated
from compass.core.schemas import hierarchy as schema
from compass.core.settings import Settings
from compass.core.utility import compass_restify

if TYPE_CHECKING:
    import requests

# TYPES_ENDPOINT_LEVELS values are meaningful values as they become the API endpoint paths
TYPES_ENDPOINT_LEVELS = Literal[
    "countries",
    "hq_sections",
    "regions",
    "country_sections",
    "counties",
    "region_sections",
    "districts",
    "county_sections",
    "groups",
    "district_sections",
    "group_sections",
]
endpoints = {i: f"/{i.replace('_', '/')}" for i in typing.get_args(TYPES_ENDPOINT_LEVELS)}


class HierarchyScraper(InterfaceAuthenticated):
    def __init__(self, session: requests.Session, member_number: int, role_number: int, jk: str):
        """Constructor for HierarchyScraper.

        takes an initialised Session object from Logon
        """
        # pylint: disable=useless-super-delegation
        # Want to keep this method for future use
        super().__init__(session, member_number, role_number, jk)

    # see CompassClient::retrieveLevel or retrieveSections in PGS\Needle php
    def get_units_from_hierarchy(
        self, parent_unit: int, level: TYPES_ENDPOINT_LEVELS
    ) -> list[Union[schema.HierarchySection, schema.HierarchyUnit, None]]:
        """Get all children of a given unit.

        If LiveData=Y is passed, the resulting JSON additionally contains:
            - (duplicated) parent id
            - the unit address
            - number of members
            - SectionType1 and SectionTypeDesc1 keys, if requesting sections data

        Args:
            parent_unit: The unit ID to get descendants from
            level: string org type, used for selecting API endpoint

        Returns:
            Mapping of unit properties to data.

            E.g.:
            {'id': ...,
             'name': '...',
             'parent_id': ...,
             'status': '...',
             'address': '...',
             'member_count': ...}

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call

        """
        # Get API endpoint from level
        level_endpoint = endpoints[level]
        # Are we requesting sections here?
        is_sections = "/sections" in level_endpoint

        result = self._post(f"{Settings.base_url}/hierarchy{level_endpoint}", json={"LiveData": "Y", "ParentID": f"{parent_unit}"})
        result_json = result.json()

        # Handle unauthorised access TODO raise???
        if result_json == {"Message": "Authorization has been denied for this request."}:
            return list()

        result_units = []
        for unit_dict in result_json:
            parsed = {
                "id": int(unit_dict["Value"]),
                "name": unit_dict["Description"],
                "parent_id": unit_dict["Parent"],
            }
            if unit_dict["Tag"]:  # TODO possible error states - what can we expect here as an invariant?
                tag = json.loads(unit_dict["Tag"])[0]
                parsed["status"] = tag["org_status"]
                parsed["address"] = tag["address"]
                parsed["member_count"] = tag["Members"]
                # Only include section_type if there is section type data
                if "SectionTypeDesc" in tag:
                    parsed["section_type"] = tag["SectionTypeDesc"]

            result_units.append(parsed)

        if is_sections:
            return pydantic.parse_obj_as(list[schema.HierarchySection], result_units)
        return pydantic.parse_obj_as(list[schema.HierarchyUnit], result_units)

    def get_members_with_roles_in_unit(
        self, unit_number: int, include_name: bool = False, include_primary_role: bool = False
    ) -> list[schema.HierarchyMember]:
        """Get details of members with roles in a given unit.

        Keys within the member_data JSON are (as at 13/01/220):
         - contact_number (membership number)
         - name (member's name)
         - visibility_status (this is meaningless as we can only see Y people)
         - address (this doesn't reliably give us postcode and is a lot of data)
         - role (This is Primary role and so only sometimes useful)

        Args:
            unit_number: Compass unit number
            include_name: include member name in returned data
            include_primary_role: include primary role in returned data

        Returns:
            A list of member records. Keys are included through args

            E.g.:
            [
                {"contact_number": ..., ...},
                ...
            ]

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call
            CompassError:
                If Compass reports that the search was invalid

        """
        keys_to_keep: tuple[str, ...] = ("contact_number",)
        if include_name:
            keys_to_keep = (*keys_to_keep, "name")
        if include_primary_role:
            keys_to_keep = (*keys_to_keep, "role")
        # Construct request data

        # It seems like the time UID value can be constant -- keeping old code in case something breaks
        # dt = datetime.datetime.now()
        # time_uid = f"{dt.hour}{dt.minute}{dt.microsecond // 1000}"
        time_uid = str(12_34_567)
        data = {"SearchType": "HIERARCHY", "OrganisationNumber": unit_number, "UI": time_uid}

        # Execute search
        # JSON data MUST be in the rather odd format of {"Key": key, "Value": value} for each (key, value) pair
        self._post(f"{Settings.base_url}/Search/Members", json=compass_restify(data))

        # Fetch results from Compass
        search_results = self._get(f"{Settings.base_url}/SearchResults.aspx")

        # Gets the compass form from the returned document
        form = html.fromstring(search_results.content).forms[0]
        del search_results

        # If the search hasn't worked the form returns an InvalidSearchError
        if form.action == "./ScoutsPortal.aspx?Invalid=SearchError":
            raise CompassError("Invalid Search")

        # Get the encoded JSON data from the HTML
        member_data_string = form.fields["ctl00$plInnerPanel_head$txt_h_Data"] or "[]"
        del form

        # parse the data and return it as a usable Python object (list)
        member_data = json.loads(member_data_string)
        return [schema.HierarchyMember(**{key: member[key] for key in keys_to_keep}) for member in member_data]



import logging

from compass.core.settings import Settings

logger = logging.getLogger("compass_interface")


def enable_debug_logging() -> None:
    import sys  # pylint: disable=import-outside-toplevel

    # Only want to import sys if needed

    logger.setLevel(logging.DEBUG)

    formatter = logging.Formatter(fmt="{asctime}.{msecs:03.0f} ({levelname}): {message}", datefmt="%Y-%m-%d %H:%M:%S", style="{")
    console = logging.StreamHandler(stream=sys.stdout)
    console.setFormatter(formatter)

    logger.addHandler(console)


if Settings.debug:
    enable_debug_logging()



class CompassError(Exception):
    pass


class CompassAuthenticationError(CompassError):
    pass


class CompassReportError(CompassError):
    pass


class CompassReportPermissionError(PermissionError, CompassError):
    pass



from __future__ import annotations

import abc
import datetime
import time
from typing import Any, Optional, TYPE_CHECKING, Union

from compass.core.logger import logger
from compass.core.settings import Settings
from compass.core.utility import compass_restify

if TYPE_CHECKING:
    from collections.abc import Mapping
    from collections.abc import MutableMapping

    import requests


class InterfaceBase:
    """Base class for interacting with the Compass server."""

    def __init__(self, session: requests.Session):
        """Constructor for InterfaceBase."""
        self.s: requests.Session = session

    def _get(
        self,
        url: Union[str, bytes],
        params: Optional[MutableMapping[str, str]] = None,
        headers: Optional[MutableMapping[str, str]] = None,
        stream: Optional[bool] = None,
        **kwargs: Any,
    ) -> requests.Response:
        """Sends a HTTP GET request.

        Pass-through method to requests.sessions.Session.get, also adding to
        the counter of total requests sent by `ci.core`.

        Args:
            url: Request URL
            params: Mapping to be sent in the query string for the request
            headers: Mapping of HTTP Headers
            stream: Whether to stream download the response content.
            kwargs: Optional arguments to requests.sessions.Session.get

        Returns:
            requests.Response object from executing the request

        Raises:
            requests.exceptions.RequestException

        """
        Settings.total_requests += 1
        return self.s.get(url, params=params, headers=headers, stream=stream, **kwargs)

    def _post(
        self,
        url: str,
        data: Optional[Mapping[str, Any]] = None,
        json: Optional[Any] = None,
        **kwargs: Any,
    ) -> requests.Response:
        """Sends a HTTP POST request.

        Pass-through method to requests.sessions.Session.post, also adding to
        the counter of total requests sent by `ci.core`.

        Args:
            url: Request URL
            data: Mapping to be sent in the query string for the request
            json: Json to send in the request body
            kwargs: Optional arguments to requests.sessions.Session.post

        Returns:
            requests.Response object from executing the request

        Raises:
            requests.exceptions.RequestException

        """
        Settings.total_requests += 1
        return self.s.post(url, data=data, json=json, **kwargs)

    def _head(self, url: str, **kwargs: Any) -> requests.Response:
        """Sends a HTTP HEAD request.

        Pass-through method to requests.sessions.Session.head, also adding to
        the counter of total requests sent by `ci.core`.

        Args:
            url: Request URL
            kwargs: Optional arguments to requests.sessions.Session.post

        Returns:
            requests.Response object from executing the request

        Raises:
            requests.exceptions.RequestException

        """
        Settings.total_requests += 1
        return self.s.head(url, **kwargs)

    def _update_headers(self, headers: dict[str, str]) -> None:
        """Update common session headers dictionary."""
        self.s.headers.update(headers)


class InterfaceAuthenticated(InterfaceBase, abc.ABC):
    def __init__(self, session: requests.Session, member_number: int, role_number: int, jk: str):
        """Constructor for InterfaceAuthenticated."""
        super(InterfaceAuthenticated, self).__init__(session)

        self.cn: int = member_number  # Contact Number
        self.mrn: int = role_number  # Member Role Number
        self.jk: str = jk  # ???? Key?  # Join Key??? SHA2-512

    def _get(
        self,
        url: str,
        params: Union[None, dict[str, str]] = None,
        headers: Optional[dict[str, str]] = None,
        stream: Optional[bool] = None,
        auth_header: bool = False,
        **kwargs: Any,
    ) -> requests.Response:
        """Sends a HTTP GET request.

        Pass-through method to requests.sessions.Session.get, also adding to
        the counter of total requests sent by `ci.core`.

        Adds custom auth_header logic for certain Compass requests
        See Scouts.js -> $.ajaxSetup -> beforeSend for details (links below)
        https://github.com/the-scouts/compass-interface/blob/master/js/Scouts.js#L1122
        https://github.com/the-scouts/compass-interface/blob/master/js/Scouts.js#L1128-L1129

        Logic in source comments is auth_header logic is only added for **AJAX**
        GET calls matching the following logic:

        if method == "GET":
            if compass_props.master.sys.safe_json is True and "system/preflight" not in url:
                return True
            elif url.lower().replace(Settings.web_service_path.lower(), "").startswith("sto_check")
                return False
            return True

        Args:
            url: Request URL
            params: Mapping to be sent in the query string for the request
            headers: Mapping of HTTP Headers
            stream: Whether to stream download the response content.
            kwargs: Optional arguments to requests.sessions.Session.get

        Returns:
            requests.Response object from executing the request

        Raises:
            requests.exceptions.RequestException

        """
        # pylint: disable=arguments-differ, too-many-arguments
        # We override the base requests.get with the custom auth logic, but
        # pylint complains that arguments differ. Also complains that we have
        # more than 5 arguments, so turn off that check too.
        if auth_header:
            if headers is None:
                headers = {}
            headers = headers | {"Auth": self._jk_hash()}

            if params is None:
                params = {}
            params = params | {
                "x1": f"{self.cn}",
                "x2": f"{self.jk}",
                "x3": f"{self.mrn}",
            }

        return super()._get(url, params=params, headers=headers, stream=stream, **kwargs)

    def _jk_hash(self) -> str:
        """Generate JK Hash needed by Compass."""
        # hash_code(f"{time.time() * 1000:.0f}")
        member_no = self.cn
        key_hash = f"{time.time() * 1000:.0f}{self.jk}{self.mrn}{member_no}"  # JK, MRN & CN are all required.
        data = compass_restify({"pKeyHash": key_hash, "pCN": member_no})
        logger.debug(f"Sending preflight data {datetime.datetime.now()}")
        self._post(f"{Settings.base_url}/System/Preflight", json=data)
        return key_hash



import datetime
import enum
from typing import Literal

from compass.core._scrapers.reports import ReportsScraper
from compass.core.errors import CompassReportError
from compass.core.logger import logger
from compass.core.logon import Logon
from compass.core.settings import Settings

TYPES_REPORTS = Literal[
    "Region Member Directory",
    "Region Appointments Report",
    "Region Permit Report",
    "Region Disclosure Report",
    "Region Training Report",
    "Region Disclosure Management Report",
]


class ReportTypes(enum.IntEnum):
    region_member_directory = 37
    region_appointments_report = 52
    region_permit_report = 72
    region_disclosure_report = 76
    region_training_report = 84
    region_disclosure_management_report = 100


class Reports:
    def __init__(self, session: Logon):
        """Constructor for Reports."""
        self._scraper = ReportsScraper(session.s, session.cn, session.mrn, session.jk)
        self.current_role: tuple[str, str] = session.current_role

    def get_report(self, report_type: TYPES_REPORTS) -> bytes:
        """Exports report as CSV from Compass.

        Exporting a report is of course surprisingly complicated. The process
        has four major steps, as follows:

        1. Get a token for generating the report from the Compass backend. This
            also validates that the report exists and that the user is
            authenticated to access it.
        2. If successful in obtaining a report token, get the initial report
            page. The token is the (relative) URL here, and the report page
            contains further needed information, such as the export URL and
            location data for a full export.
            (Compass does not include all organisational units in reports by
            default, and to export all data for a given unit and downwards, we
            need to add in these missing/unset levels manually).
        3. We update report configuration data (sent as form data), and check
            that we are not in an error state.
        4. We extract the export URL, download the content and save to disk.

        Pitfalls to be aware of in this process include that:
        - Compass checks user-agent headers in some parts of the process
            (TODO pinpoint which exactly)
        - There is a ten (10) minute default soft-timeout, which may run out
            before a report download has finished
        - If a requested report is too large, Compass can simply give up, often
            with an `OutOfMemory` error or similar

        Returns:
            Report output, as a bytes-encoded object.

        Raises:
            CompassReportError:
                - If the user passes an invalid report type
                - If Compass returns a JSON error
                - If there is an error updating the form data
            CompassReportPermissionError:
                If the user does not have permission to run the report
            requests.HTTPError:
                If there is an error in the transport layer, or if Compass
                reports a HTTP 5XX status code

        """
        # Get token for report type & role running said report:
        try:
            # report_type is given as `Title Case` with spaces, enum keys are in `snake_case`
            rt_key = report_type.lower().replace(" ", "_")
            run_report_url = self._scraper.get_report_token(ReportTypes[rt_key].value, self._scraper.mrn)
        except KeyError:
            # enum keys are in `snake_case`, output types as `Title Case` with spaces
            types = [rt.name.title().replace("_", " ") for rt in ReportTypes]
            raise CompassReportError(f"{report_type} is not a valid report type. Existing report types are {types}") from None

        # Get initial reports page, for export URL and config:
        report_page = self._scraper.get_report_page(run_report_url)

        # Update form data & set location selection:
        self._scraper.update_form_data(report_page, f"{Settings.base_url}/{run_report_url}")

        # Export the report:
        logger.info("Exporting report")
        export_url_path, export_url_params = self._scraper.get_report_export_url(report_page.decode("UTF-8"))

        # TODO Debug check
        time_string = datetime.datetime.now().replace(microsecond=0).isoformat().replace(":", "-")  # colons are illegal on windows
        filename = f"{time_string} - {self._scraper.cn} ({' - '.join(self.current_role)}).csv"

        # start = time.time()
        # TODO TRAINING REPORT ETC.
        # # TODO REPORT BODY HAS KEEP ALIVE URL KeepAliveUrl
        # p = PeriodicTimer(15, lambda: self.report_keep_alive(self.session, report_page.text))
        # self.session.sto_thread.start()
        # p.start()
        # # ska_url = self.report_keep_alive(self.session, report_page.text)
        # try:
        #     self.download_report(self.session, f"{Settings.base_url}/{export_url_path}", export_url_params, filename, )  # ska_url
        # except (ConnectionResetError, requests.ConnectionError):
        #     logger.info(f"Stopped at {datetime.datetime.now()}")
        #     p.cancel()
        #     self.session.sto_thread.cancel()
        #     raise
        # logger.debug(f"Exporting took {time.time() -start}s")

        csv_export = self._scraper.download_report_normal(f"{Settings.base_url}/{export_url_path}", export_url_params, filename)

        return csv_export



import pydantic


class _SettingsModel(pydantic.BaseSettings):
    base_domain: str = "compass.scouts.org.uk"
    base_url: pydantic.HttpUrl = f"https://{base_domain}"
    date_format: str = "%d %B %Y"  # dd Month YYYY
    org_number: int = 10000001
    total_requests: int = 0
    wcf_json_endpoint: str = "/JSon.svc"  # Windows communication foundation JSON service endpoint
    web_service_path: pydantic.HttpUrl = base_url + wcf_json_endpoint
    debug: bool = False

    class Config:  # noqa: D106
        case_sensitive = False  # this is the default, but mark for clarity.
        env_prefix = "CI_"  # env variables named `REDIS_HOST` etc


Settings = _SettingsModel()



from __future__ import annotations

import datetime
from pathlib import Path
import re
import time
from typing import Optional, TYPE_CHECKING
import urllib.parse

from lxml import html

from compass.core import utility
from compass.core.errors import CompassReportError
from compass.core.errors import CompassReportPermissionError
from compass.core.interface_base import InterfaceAuthenticated
from compass.core.logger import logger
from compass.core.settings import Settings

if TYPE_CHECKING:
    import requests


class ReportsScraper(InterfaceAuthenticated):
    def __init__(self, session: requests.Session, member_number: int, role_number: int, jk: str):
        """Constructor for ReportsScraper.

        takes an initialised Session object from Logon
        """
        # pylint: disable=useless-super-delegation
        # Want to keep this method for future use
        super().__init__(session, member_number, role_number, jk)

    def get_report_token(self, report_number: int, role_number: int) -> str:
        params = {
            "pReportNumber": str(report_number),
            "pMemberRoleNumber": str(role_number),
        }
        logger.debug("Getting report token")
        response = self._get(f"{Settings.web_service_path}/ReportToken", auth_header=True, params=params)
        response.raise_for_status()

        report_token_uri = str(response.json().get("d"))
        if report_token_uri not in {"-1", "-2", "-3", "-4"}:
            return report_token_uri
        elif report_token_uri in {"-2", "-3"}:
            raise CompassReportError("Report aborted: Report No Longer Available")
        elif report_token_uri == "-4":
            raise CompassReportPermissionError("Report aborted: USER DOES NOT HAVE PERMISSION")

        raise CompassReportError("Report aborted")

    @staticmethod
    def get_report_export_url(report_page: str, filename: Optional[str] = None) -> tuple[str, dict[str, str]]:
        full_url = re.search(r'"ExportUrlBase":"(.*?)"', report_page).group(1).encode().decode("unicode-escape")
        fragments = urllib.parse.urlparse(full_url)
        export_url_path = fragments.path[1:]  # strip leading `/`
        report_export_url_data = dict(urllib.parse.parse_qsl(fragments.query, keep_blank_values=True))
        report_export_url_data["Format"] = "CSV"
        if filename is not None:
            report_export_url_data["FileName"] = filename

        return export_url_path, report_export_url_data

    def get_report_page(self, run_report_url: str) -> bytes:
        # TODO what breaks if we don't update user-agent?
        # Compass does user-agent sniffing in reports!!!
        self._update_headers({"User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"})

        # Get initial reports page, for export URL and config.
        logger.info("Generating report")
        report_page = self._get(f"{Settings.base_url}/{run_report_url}")

        return report_page.content

    def update_form_data(self, report_page: bytes, run_report: str) -> None:
        form = html.fromstring(report_page).forms[0]
        elements = {el.name: el.value for el in form.inputs if el.get("type") not in {"checkbox", "image"}}

        # Table Controls: table#ParametersGridReportViewer1_ctl04
        # ReportViewer1$ctl04$ctl03$ddValue - Region/County(District) Label
        # ReportViewer1$ctl04$ctl05$txtValue - County Label
        # ReportViewer1$ctl04$ctl07$txtValue - District Label
        # ReportViewer1$ctl04$ctl09$txtValue - Role Types (Status)
        # ReportViewer1$ctl04$ctl15$txtValue - Columns Label

        # ReportViewer1_ctl04_ctl07_divDropDown - Districts
        # ReportViewer1_ctl04_ctl05_divDropDown - Counties
        # ReportViewer1_ctl04_ctl09_divDropDown - Role Types
        # ReportViewer1_ctl04_ctl15_divDropDown - Columns

        form_data = {
            "__VIEWSTATE": elements["__VIEWSTATE"],
            "ReportViewer1$ctl04$ctl05$txtValue": "Regional Roles",
            "ReportViewer1$ctl04$ctl05$divDropDown$ctl01$HiddenIndices": "0",
        }

        # districts = tree.xpath("//div[@id='ReportViewer1_ctl04_ctl07_divDropDown']//label/text()")
        # list_districts = [unicodedata.normalize("NFKD", d) for d in districts if "select all" not in d.lower()]
        # numbered_districts = {str(i): d for i, d in enumerate(list_districts)}
        # all_districts = ", ".join(numbered_districts.values())
        # all_districts_indices = ",".join(numbered_districts.keys())
        #
        # counties = tree.xpath("//div[@id='ReportViewer1_ctl04_ctl05_divDropDown']//label/text()")
        # list_counties = [unicodedata.normalize("NFKD", c) for c in counties if "select all" not in c.lower()]
        # numbered_counties = {str(i): c for i, c in enumerate(list_counties)}
        # all_counties = ", ".join(numbered_counties.values())
        # all_counties_indices = ",".join(numbered_counties.keys())
        #
        # form_data = {
        #     "__VIEWSTATE": elements["__VIEWSTATE"],
        #     "ReportViewer1$ctl04$ctl05$txtValue": all_counties,
        #     "ReportViewer1$ctl04$ctl05$divDropDown$ctl01$HiddenIndices": all_counties_indices,
        #     "ReportViewer1$ctl04$ctl07$txtValue": all_districts,
        #     "ReportViewer1$ctl04$ctl07$divDropDown$ctl01$HiddenIndices": all_districts_indices,
        # }

        # Including MicrosoftAJAX: Delta=true reduces size by ~1kb but increases time by 0.01s.
        # In reality we don't care about the output of this POST, just that it doesn't fail
        report = self._post(run_report, data=form_data, headers={"X-MicrosoftAjax": "Delta=true"})
        report.raise_for_status()

        # Check error state
        if "compass.scouts.org.uk%2fError.aspx|" in report.text:
            raise CompassReportError("Compass Error!")

    def report_keep_alive(self, report_page: str) -> str:
        logger.info(f"Extending Report Session {datetime.datetime.now()}")
        keep_alive = re.search(r'"KeepAliveUrl":"(.*?)"', report_page).group(1).encode().decode("unicode-escape")
        response = self._post(f"{Settings.base_url}{keep_alive}")  # NoQA: F841

        return keep_alive  # response

    def download_report_streaming(self, url: str, params: dict[str, str], filename: str) -> None:
        with self._get(url, params=params, stream=True) as r:
            r.raise_for_status()
            with utility.filesystem_guard("Unable to write report export"), open(filename, "wb") as f:  # TODO swap `with` stmts?
                for chunk in r.iter_content(chunk_size=1024 ** 2):  # Chunk size == 1MiB
                    f.write(chunk)

    def download_report_normal(self, url: str, params: dict[str, str], filename: str) -> bytes:
        start = time.time()
        csv_export = self._get(url, params=params)
        logger.debug(f"Exporting took {time.time() - start}s")
        logger.info("Saving report")
        with utility.filesystem_guard("Unable to write report export"):
            Path(filename).write_bytes(csv_export.content)  # TODO Debug check
        logger.info("Report Saved")

        logger.debug(len(csv_export.content))

        return csv_export.content



import contextlib
import ctypes
import datetime
import functools
import threading
from typing import Any, Callable, Optional

import pydantic

from compass.core.logger import logger


def hash_code(text: str) -> int:
    """Implements Java's hashCode in Python.

    Ref: https://stackoverflow.com/a/8831937
    """
    return functools.reduce(lambda code, char: ctypes.c_int32(31 * code + ord(char)).value, list(text), 0)


def compass_restify(data: dict[str, Any]) -> list[dict[str, str]]:
    """Format a dictionary of key-value pairs into the correct format for Compass.

    It seems that JSON data MUST be in the rather odd format of {"Key": key, "Value": value} for each (key, value) pair.
    """
    return [{"Key": f"{k}", "Value": f"{v}"} for k, v in data.items()]


def maybe_int(value: Any) -> Optional[int]:
    """Casts value to int or None."""
    try:
        return int(value)
    except ValueError:
        return None


def parse(date_time_str: str) -> Optional[datetime.datetime]:
    if not date_time_str:
        return None
    else:
        try:
            return datetime.datetime.strptime(date_time_str, "%d %B %Y")  # e.g. 01 January 2000
        except ValueError:
            return datetime.datetime.strptime(date_time_str, "%d %b %Y")  # e.g. 01 Jan 2000


@contextlib.contextmanager
def filesystem_guard(msg: str):
    try:
        yield
    except IOError as err:
        logger.error(f"{msg}: {err.errno} - {err.strerror}")


@contextlib.contextmanager
def validation_errors_logging(id_value: int, name: str = "Member No"):
    try:
        yield
    except pydantic.ValidationError as err:
        logger.error(f"Parsing Error! {name}: {id_value}")
        raise err


class PeriodicTimer:
    def __init__(self, interval: float, callback: Callable[..., Any]):
        """Constructor for PeriodicTimer."""
        self.interval = interval

        @functools.wraps(callback)
        def wrapper(*args: Any, **kwargs: Any) -> None:
            result = callback(*args, **kwargs)
            if result is not None:
                self.thread = threading.Timer(self.interval, self.callback)
                self.thread.start()

        self.callback = wrapper
        self.thread: threading.Timer = threading.Timer(0.0, self.callback)

    def start(self) -> "PeriodicTimer":
        self.thread.start()
        return self

    def cancel(self) -> "PeriodicTimer":
        self.thread.cancel()
        return self


#
# def set_interval(interval: int):
#     def decorator(function: Callable):
#         def wrapper(*args, **kwargs):
#             stopped = threading.Event()
#
#             def loop():  # executed in another thread
#                 while not stopped.wait(interval):  # until stopped
#                     function(*args, **kwargs)
#
#             t = threading.Thread(target=loop)
#             t.daemon = True  # stop if the program exits
#             t.start()
#             return stopped
#         return wrapper
#     return decorator
#
#
# import asyncio
#
#
# async def periodic(n):
#     print('periodic')  # run immediately
#     while await asyncio.sleep(n, result=True):
#         print('periodic')
#
# loop = asyncio.get_event_loop()
# task = loop.create_task(periodic(0.5))
# loop.call_later(5, task.cancel)
# with contextlib.suppress(asyncio.CancelledError):
#     loop.run_until_complete(task)
#
#
# # wrapper:
# def periodic(period: int):
#     def scheduler(function: Callable):
#         async def wrapper(*args, **kwargs):
#             asyncio.create_task(function(*args, **kwargs))  # run immediately
#             while await asyncio.sleep(period, result=True):
#                 asyncio.create_task(function(*args, **kwargs))
#         return wrapper
#     return scheduler



from __future__ import annotations

import datetime
from typing import Any, Generic, Literal, Optional, TYPE_CHECKING, TypeVar, Union
import warnings

import phonenumbers
import pydantic
from pydantic import generics

# Must use typing.Dict etc for generics not native as of pydantic 1.7.3
from typing import List, Dict  # isort: skip

if TYPE_CHECKING:
    from collections.abc import Iterator

DataT = TypeVar("DataT")

TYPES_ROLE_STATUS = Literal["Cancelled", "Closed", "Full", "Pre provisional", "Provisional"]
TYPES_SEX = Literal["Male", "Female", "Unknown"]
TYPES_ETHNICITY = Literal[
    "1.English/Welsh/Scottish/Northern Irish/British",
    "2.Irish",
    "3.Gypsy or Irish Traveller",
    "4.Any other White background",
    "5.White and Black Caribbean",
    "6.White and Black African",
    "7.White and Asian",
    "8.Any other mixed or Multiple ethic group",
    "9.Indian",
    "10.Pakistani",
    "11.Bangladeshi",
    "12.Chinese",
    "13.Any other Asian Background",
    "14.African",
    "15.Caribbean",
    "16.Any other Black/African/Caribbean background",
    "17.Arab",
    "18.Other",
    "19.Prefer not to say",
]
TYPES_RELIGION = Union[
    Literal[
        "Buddhist",
        "Christian (including all Christian denominations)",
        "Hindu",
        "Jewish",
        "Muslim",
        "Any other religion (please specify)",
        "No religion",
        "Prefer not to say",
    ],
    pydantic.constr(regex=r"^Christian.*"),  # NoQA F722 (https://stackoverflow.com/a/64917499)
    pydantic.constr(regex=r"^Any other religion.*"),  # NoQA F722 (https://stackoverflow.com/a/64917499)
    pydantic.constr(regex=r"^No religion.*"),  # NoQA F722 (https://stackoverflow.com/a/64917499)
]
TYPES_OCCUPATION = Union[
    Literal[
        "Employed",
        "Unemployed",
        "Retired (whether receiving a pension or not)",
        "Student",
        "Long term sick or disabled",
        "Looking after home of family",
        "Other",
    ],
    pydantic.constr(regex=r"^Employed.*"),  # NoQA F722 (https://stackoverflow.com/a/64917499)
    pydantic.constr(regex=r"^Unemployed.*"),  # NoQA F722 (https://stackoverflow.com/a/64917499)
    pydantic.constr(regex=r"^Retired.*"),  # NoQA F722 (https://stackoverflow.com/a/64917499)
    pydantic.constr(regex=r"^Student.*"),  # NoQA F722 (https://stackoverflow.com/a/64917499)
    pydantic.constr(regex=r"^Long term sick or disabled.*"),  # NoQA F722 (https://stackoverflow.com/a/64917499)
    pydantic.constr(regex=r"^Looking after home of family.*"),  # NoQA F722 (https://stackoverflow.com/a/64917499)
    pydantic.constr(regex=r"^Other.*"),  # NoQA F722 (https://stackoverflow.com/a/64917499)
]
TYPES_ROLE_CLASS = Literal[
    "Administrator",
    "Advisor",
    "Assessor",
    "Co-ordinator",
    "Commissioner",
    "Committee",
    "Helper",
    "Honorary",
    "Leader",
    "Manager",
    "Secretary",
    "Staff",
    "System Role",
    "Supporter",
    "Trainer",
    "Default role class",
]
TYPES_REFERENCES = Literal[
    "Not Complete",
    "Not Required",
    "References Requested",
    "References Satisfactory",
    "References Unsatisfactory",
]
TYPES_DISCLOSURES_APPOINTMENT = Literal[
    "Application submitted - in progress",
    "Disclosure Issued",
    "ID check required",
    "No Disclosure",
]  # Disclosure statuses in role details popup
TYPES_PERMIT = Literal[
    "Archery",
    "Bell Boating",
    "Canoeing",
    "Caving",
    "Climbing and Abseiling",
    "Dinghy Sailing",
    "Dragon Boating",
    "Hill Walking",
    "Hovercrafting",
    "Ice Climbing",
    "Kayaking",
    "Keelboating",
    "Kite Surfing",
    "Mine Exploration",
    "Motor Cruising",
    "Mountain Biking",
    "Narrow Boating",
    "Nights Away",
    "Personal Watercraft (Jet Ski)",
    "Power Boating",
    "Pulling",
    "Rafting (Traditional)",
    "Rowing and Sculling",
    "Scuba Diving",
    "Snorkelling",
    "Snowsports",
    "Water Skiing",
    "White Water Rafting",
    "Windsurfing",
    "Yachting",
]
TYPES_PERMIT_CATEGORIES = Literal[
    "B1 Waters",
    "B2 Waters",
    "B2+ Waters",
    "B3 Waters",
    "C Waters",
    "Campsite",
    "Greenfield",
    "Indoor",
    "Lightweight Expedition",
    "Open Inland B1 Waters",
    "Open Inland B2 Waters",
    "Open Inland B3 Waters",
    "River B1 Waters",
    "Sea B1 Waters",
    "Sea B2 Waters",
    "Surf B2 Waters",
    "With Compound Bows",
    "Without Compound Bows",
]
TYPES_PERMIT_TYPE = Literal["Leadership", "Supervisory"]
TYPES_AWARD_TYPE = Literal[
    "Chief Scout's 5 years Service Award",
    "Chief Scout's 10 years Service Award",
    "Chief Scout's 15 years Service Award",
    "Chief Scout's 20 years Service Award",
    "Chief Scout's 25 years Service Award",
    "Chief Scout's 30 years Service Award",
    "Chief Scout's 40 years Service Award",
    "Chief Scout's 50 years Service Award",
    "Chief Scout's 60 years Service Award",
    "Chief Scout's 70 years Service Award",
    # Local / less formal awards
    "Commissioners Commendation",
    # Formal awards process - lower good service
    "Chief Scout's Commendation for Good Service",
    "Award for Merit",
    "Bar to the Award for Merit",
    "Silver Acorn",
    # Formal awards process - higher good service
    "Bar to the Silver Acorn",
    "Silver Wolf",
    # ???
    "Medal of Merit",
    "Bar to the Medal of Merit",
]
TYPES_DISCLOSURE_PROVIDERS = Literal["Atlantic Data", "DBS Paper Application", "Disclosure Scotland", "Other"]
TYPES_DISCLOSURE_STATUSES = Literal[
    "Application Withdrawn",
    "Application submitted - in progress",
    "Disclosure Expired",
    "Disclosure Issued",
    "Expired",
    "ID check required",
    "ID selection required",
]  # Disclosure statuses in disclosures tab


class MemberGenericDict(generics.GenericModel, Generic[DataT]):
    __root__: Dict[int, DataT]  # Must use typing.Dict not dict as of pydantic 1.7.3

    def __iter__(self) -> Iterator[tuple[int, DataT]]:
        """Iterate over model items."""
        yield from self.__root__.items()

    def __getitem__(self, item: int) -> DataT:
        """Get item by key."""
        return self.__root__[item]

    def __len__(self) -> int:
        """Get number of items."""
        return len(self.__root__)

    def items(self) -> Iterator[tuple[int, DataT]]:
        yield from iter(self)


class MemberGenericList(generics.GenericModel, Generic[DataT]):
    __root__: List[DataT]

    def __iter__(self) -> Iterator[DataT]:
        """Iterate over model items."""
        yield from self.__root__

    def __getitem__(self, item: int) -> DataT:
        """Get item by index."""
        return self.__root__[item]

    def __len__(self) -> int:
        """Get number of items."""
        return len(self.__root__)


class MemberBase(pydantic.BaseModel):
    membership_number: int


class MemberRoleBase(pydantic.BaseModel):
    # Role details
    role_number: int
    role_title: str  # role_name
    role_start: datetime.date
    role_status: TYPES_ROLE_STATUS
    review_date: Optional[datetime.date] = None


# Personal Details Tab
class MemberDetails(MemberBase):
    # Name
    name: str
    known_as: str
    forenames: Optional[str] = None
    surname: Optional[str] = None

    # Core personal information
    birth_date: Optional[datetime.date] = None
    sex: Optional[TYPES_SEX] = None
    nationality: Optional[str] = None  # literal? Big list...!
    ethnicity: Optional[TYPES_ETHNICITY] = None
    religion: Optional[TYPES_RELIGION] = None
    occupation: Optional[TYPES_OCCUPATION] = None
    join_date: Optional[datetime.date] = None

    # Contact Details
    main_phone: Optional[str] = None
    main_email: Optional[pydantic.EmailStr] = None
    address: Optional[str] = None  # this is the unmodified address string
    country: Optional[str] = None
    postcode: Optional[str] = None
    county: Optional[str] = None
    town: Optional[str] = None
    street: Optional[str] = None

    # Additional / miscellaneous details
    # TODO - potential disabilities, qualifications, hobbies sections

    @pydantic.validator("main_phone")
    def check_phone_number(cls, v: Optional[str], values: dict[str, Any]) -> Optional[str]:
        if v is None or not v or v == "0" or len(v) < 2:
            return None

        try:
            n = phonenumbers.parse(v, "GB")
        except phonenumbers.NumberParseException as err:
            cn = values["membership_number"]
            raise ValueError(f"Member No {cn}: phone number {v} is not valid!") from err

        if not phonenumbers.is_valid_number(n):
            cn = values["membership_number"]
            warnings.warn(f"Member No {cn}: phone number {v} is not valid!", RuntimeWarning)

        fmt = phonenumbers.PhoneNumberFormat.NATIONAL if n.country_code == 44 else phonenumbers.PhoneNumberFormat.INTERNATIONAL
        return str(phonenumbers.format_number(n, fmt))


# Roles Tab (Main List - item)
class MemberRoleCore(MemberBase, MemberRoleBase):
    # Role details
    role_end: Optional[datetime.date] = None
    role_class: TYPES_ROLE_CLASS
    role_type: Optional[str] = None  # TODO literal

    # Location details
    location_id: Optional[int] = None
    location_name: Optional[str] = None

    # Is the role details popup accessible?
    can_view_details: bool


# Roles Tab (Main List - collection)
MemberRolesDict = MemberGenericDict[MemberRoleCore]


# Roles Tab (Role Detail Popup - Main)
class MemberRoleDetail(MemberBase, MemberRoleBase):
    organisation_level: str
    birth_date: Optional[datetime.date]
    name: str

    # Approval Process
    line_manager_number: Optional[int] = None
    line_manager: Optional[str] = None

    # Approval information
    ce_check: Optional[datetime.date]  # Optional for Closed roles - e.g. #499, role closed 1976, or if Pending
    disclosure_check: Optional[TYPES_DISCLOSURES_APPOINTMENT]
    disclosure_date: Optional[datetime.date]
    references: Optional[TYPES_REFERENCES] = None
    appointment_panel_approval: Optional[Literal["NC", "NR", "S", "U"]] = None
    commissioner_approval: Optional[Literal["NC", "NR", "RR", "S", "U"]] = None
    committee_approval: Optional[Literal["NC", "S", "U"]] = None


# Roles Tab (Role Detail Popup - Hierarchy)
class MemberRoleHierarchy(pydantic.BaseModel):
    # Role location
    organisation: str
    country: Optional[str] = None
    region: Optional[str] = None
    county: Optional[str] = None
    district: Optional[str] = None
    group: Optional[str] = None
    section: Optional[str] = None


# Roles Tab (Role Detail Popup - Getting Started)
class MemberRoleGettingStartedModule(pydantic.BaseModel):
    # name: str  # Module name  # needed?? as saved under key
    validated: Optional[datetime.date] = None
    validated_by: Optional[str] = None


# Roles Tab (Role Detail Popup - Getting Started)
class MemberRoleGettingStarted(pydantic.BaseModel):
    # Getting Started Training
    module_01: Optional[MemberRoleGettingStartedModule] = None
    trustee_intro: Optional[MemberRoleGettingStartedModule] = None
    module_02: Optional[MemberRoleGettingStartedModule] = None
    module_03: Optional[MemberRoleGettingStartedModule] = None
    module_04: Optional[MemberRoleGettingStartedModule] = None
    gdpr: Optional[MemberRoleGettingStartedModule] = None
    # training_completion_date: Optional[str] = None  # TODO where is this from???!!!


# Roles Tab (Role Detail Popup - All)
class MemberRolePopup(pydantic.BaseModel):
    hierarchy: MemberRoleHierarchy
    details: MemberRoleDetail
    getting_started: MemberRoleGettingStarted


# Training Tab (Role)
class MemberTrainingRole(MemberRoleBase):
    # Location details
    location: Optional[str] = None

    # Training Advisor details
    ta_data: Optional[str]
    ta_number: Optional[int]
    ta_name: Optional[str]

    # Completion details
    completion: Optional[str]
    completion_type: Optional[str]
    completion_date: Optional[datetime.date]
    wood_badge_number: Optional[int]  # WB_1234567


# Training Tab (PLP)
class MemberTrainingPLP(pydantic.BaseModel):
    pk: int
    module_id: int
    code: str  # TODO literal
    name: str  # TODO literal

    # Learning details
    learning_required: Optional[bool]
    learning_method: Optional[str]
    learning_completed: Optional[datetime.date]
    learning_date: Optional[datetime.date]

    # Validation details
    validated_membership_number: Optional[int]
    validated_name: Optional[str]
    validated_date: Optional[datetime.date]


# Training Tab (OGL - item)
class MemberMOGL(pydantic.BaseModel):
    completed_date: Optional[datetime.date]
    renewal_date: Optional[datetime.date]  # GDPR has no renewal date...??


# Training Tab (OGL - all)
class MemberMOGLList(pydantic.BaseModel):
    safety: MemberMOGL
    safeguarding: MemberMOGL
    first_aid: MemberMOGL
    gdpr: MemberMOGL


# Training Tab
class MemberTrainingTab(pydantic.BaseModel):
    roles: dict[int, MemberTrainingRole]
    plps: dict[int, list[MemberTrainingPLP]]
    mandatory: MemberMOGLList


# Permits Tab
class MemberPermit(MemberBase):
    permit_type: TYPES_PERMIT
    category: str
    type: TYPES_PERMIT_TYPE
    restrictions: str
    expires: Optional[datetime.date]
    status: Literal["PERM_OK", "PERM_EXP", "PERM_REV"]


# Permits Tab - collection
MemberPermitsList = MemberGenericList[MemberPermit]


class MemberAward(MemberBase):
    type: TYPES_AWARD_TYPE
    location: Optional[str]
    date: datetime.date


class MemberDisclosure(MemberBase):
    # TODO Scot, Wales, NI, BSO, Branches, Channel Islands
    country: Optional[Literal["England & Wales", "Scotland", "The Scout Association"]]
    provider: TYPES_DISCLOSURE_PROVIDERS
    type: Literal["Enhanced with Barring"]
    number: Union[Optional[int], pydantic.constr(regex=r"^\d{7}R$")]  # NoQA: F722 # If Application Withdrawn, no disclosure number
    issuer: Optional[TYPES_DISCLOSURE_PROVIDERS]
    issue_date: Optional[datetime.date]  # If Application Withdrawn, maybe no issue date
    status: TYPES_DISCLOSURE_STATUSES
    expiry_date: Optional[datetime.date]  # If Application Withdrawn, no expiry date



from __future__ import annotations

import datetime
from typing import Literal, Optional, TYPE_CHECKING, Union
import urllib.parse

from lxml import html
import requests

from compass.core import schemas
from compass.core.errors import CompassAuthenticationError
from compass.core.errors import CompassError
from compass.core.interface_base import InterfaceAuthenticated
from compass.core.interface_base import InterfaceBase
from compass.core.logger import logger
import compass.core.schemas.logon as schema
from compass.core.settings import Settings
from compass.core.utility import PeriodicTimer

if TYPE_CHECKING:
    from collections.abc import Iterator

TYPES_UNIT_LEVELS = Literal["Group", "District", "County", "Region", "Country", "Organisation"]
TYPES_STO = Literal[None, "0", "5", "X"]
TYPES_ROLES_DICT = dict[int, tuple[str, str]]
TYPES_ROLE = tuple[str, str]


def login(username: str, password: str, /, *, role: Optional[str] = None, location: Optional[str] = None) -> Logon:
    """Log in to compass, return a compass.logon.Logon object.

    This function is provided as a convenient interface to the logon module.
    """
    return Logon.from_logon((username, password), role, location)


class Logon(InterfaceAuthenticated):
    """Create connection to Compass and authenticate. Holds session state.

    Logon flow is:
    1. Create a persistent state object (Session) to hold headers, cookies etc.
      a. Get ASP.NET session cookie from Compass. (HTTP request #1)
    2. Post login data to Compass. (HTTP request #2)
    3. Get sample page from Compass. (HTTP request #3)
    4. Verify login was successful
      a. Create dict of compass internal variables, e.g. Master.User.CN (POST_CTRL).
      b. Create dict of user's role title to internal role number
      c. Update data for authorisation headers for requests to Compass.
    IF Role title specified:
    5. Post new role number to Compass (HTTP request #4)
    6. Get sample page from Compass. (HTTP request #5)
    7. Get currently active role number, and check this equals the requested role number.
      a. Create dict of compass internal variables, e.g. Master.User.CN (POST_CTRL).
      b. Create dict of user's role title to internal role number
      c. Update data for authorisation headers for requests to Compass.

    """

    def __init__(
        self,
        *,
        session: requests.Session,
        compass_props: Optional[schema.CompassProps] = None,
        roles_dict: Optional[TYPES_ROLES_DICT] = None,
        current_role: Optional[TYPES_ROLE] = None,
    ):
        """Constructor for Logon."""
        self.compass_props: schema.CompassProps = compass_props
        self.roles_dict: TYPES_ROLES_DICT = roles_dict or {}
        self.current_role: TYPES_ROLE = current_role or ("", "")

        self.sto_thread = PeriodicTimer(150, self._extend_session_timeout)
        # self.sto_thread.start()

        # Set these last, treat as immutable after we leave init. Role can
        # theoretically change, but this is not supported behaviour.

        member_number = self.compass_props.master.user.cn  # Contact Number
        role_number = self.compass_props.master.user.mrn  # Member Role Number
        jk = self.compass_props.master.user.jk  # ???? Key?  # Join Key??? SHA2-512

        self._asp_net_id: str = session.cookies["ASP.NET_SessionId"]
        self._session_id: str = self.compass_props.master.sys.session_id

        # Finally, call super
        super().__init__(session, member_number, role_number, jk)

    @classmethod
    def from_logon(
        cls,
        credentials: Optional[tuple[str, str]] = None,
        role_to_use: Optional[str] = None,
        role_location: Optional[str] = None,
    ) -> Logon:
        """Initialise a Logon object with login information.

        Args:
            credentials: username and password
            role_to_use: Compass role to use
            role_location: Role location, if multiple identical role titles

        Returns:
            Initialised Logon object

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call
            CompassError:
                If initial connection to Compass fails
            CompassAuthenticationError:
                If authentication with Compass fails

        """
        worker = LogonCore()
        session = worker.s

        # Log in and try to confirm success
        _response, props, roles = worker.logon_remote(credentials)

        logon = cls(
            session=session,
            compass_props=props,
            roles_dict=roles,
            current_role=roles[props.master.user.mrn],  # Set explicitly as work done in worker
        )

        if role_to_use is not None:
            # Session contains updated auth headers from role change
            logon._change_role(session, role_to_use, role_location)  # pylint: disable=protected-access

        return logon

    @classmethod
    def from_session(cls, asp_net_id: str, user_props: dict[str, Union[str, int]], current_role: TYPES_ROLE) -> Logon:
        """Initialise a Logon object with stored data.

        This method  is used to avoid logging in many times, by enabling reuse
        of an existing sever-side session in Compass. It is used by the main
        compass-interface web API.

        Args:
            asp_net_id: ASP.NET Session ID, from cookie
            user_props: Compass master.sys.user properties
            current_role: Role used by initialised session

        Returns:
            Initialised Logon object

        Raises:
            requests.exceptions.RequestException:
                For errors while executing the HTTP call

        """
        session = requests.Session()
        session.cookies.set("ASP.NET_SessionId", asp_net_id, domain=Settings.base_domain)

        logon = cls(
            session=session,
            compass_props=schema.CompassProps(**{"master": {"user": dict(user_props)}}),
            current_role=current_role,
        )

        LogonCore(session=session).update_auth_headers(logon.cn, logon.mrn, logon._session_id)  # pylint: disable=protected-access

        return logon

    def __repr__(self) -> str:
        """String representation of the Logon class."""
        return f"{self.__class__} Compass ID: {self.cn} ({' - '.join(self.current_role)})"

    # properties/accessors code:

    @property
    def hierarchy(self) -> schemas.hierarchy.HierarchyLevel:
        unit_number = self.compass_props.master.user.on  # Organisation Number
        unit_level = self.compass_props.master.user.lvl  # Level
        level_map = {
            "ORG": "Organisation",
            # "ORST": "Organisation Sections",
            "CNTR": "Country",
            # "CNST": "Country Sections",
            "REG": "Region",
            # "RGST": "Regional Sections",
            "CNTY": "County",  # Also Area/Scot Reg/Branch
            # "CTST": "County Sections",  # Also Area/Scot Reg/Branch
            "DIST": "District",
            # "DTST": "District Sections",
            "SGRP": "Group",
            # "SGST": "Group Sections",
        }

        return schemas.hierarchy.HierarchyLevel(id=unit_number, level=level_map[unit_level])

    # Timeout code:

    def _extend_session_timeout(self, sto: TYPES_STO = "0") -> requests.Response:
        # Session time out. 4 values: None (normal), 0 (STO prompt) 5 (Extension, arbitrary constant) X (Hard limit)
        logger.debug(f"Extending session length {datetime.datetime.now()}")
        # TODO check STO.js etc for what happens when STO is None/undefined
        return self._get(f"{Settings.web_service_path}/STO_CHK", auth_header=True, params={"pExtend": sto})

    def _change_role(self, session: requests.Session, new_role: str, location: Optional[str] = None) -> Logon:
        """Returns new Logon object with new role.

        If the user has multiple roles with the same role title, the first is used,
        unless the location parameter is set, where the location is exactly matched.
        """
        logger.info("Changing role")

        new_role = new_role.strip()

        worker = LogonCore(session=session)

        # If we don't have the roles dict, generate it.
        if not self.roles_dict:
            _props, self.roles_dict = worker.check_login()

        # Change role to the specified role number
        if location is not None:
            location = location.strip()
            member_role_number = next(num for num, name in self.roles_dict.items() if name == (new_role, location))
        else:
            member_role_number = next(num for num, name in self.roles_dict.items() if name[0] == new_role)

        response = session.post(f"{Settings.base_url}/API/ChangeRole", json={"MRN": member_role_number})  # b"false"
        Settings.total_requests += 1
        logger.debug(f"Compass ChangeRole call returned: {response.json()}")

        # Confirm Compass is reporting the changed role number, update auth headers
        self.compass_props, self.roles_dict = worker.check_login(check_role_number=member_role_number)
        self.current_role = self.roles_dict[member_role_number]  # Set explicitly as work done in worker

        logger.info(f"Role updated successfully! Role is now {self.current_role[0]} ({self.current_role[1]}).")

        return self


class LogonCore(InterfaceBase):
    def __init__(self, session: Optional[requests.Session] = None):
        """Initialise InterfaceBase with a session."""
        if session is None:
            super().__init__(self._create_session())
        else:
            super().__init__(session)

    @staticmethod
    def _create_session() -> requests.Session:
        """Create a session and get ASP.Net Session ID cookie from the compass server."""
        session = requests.Session()

        session.head(f"{Settings.base_url}/")  # use .head() as only headers needed to grab session cookie
        Settings.total_requests += 1

        if not session.cookies:
            raise CompassError(
                "Could not create a session with Compass. Please check that this programme can "
                "access Compass (including firewalls etc.) and that Compass is currently online. "
            )

        return session

    def logon_remote(self, auth: tuple[str, str]) -> tuple[requests.Response, schema.CompassProps, TYPES_ROLES_DICT]:
        """Log in to Compass and confirm success."""
        # Referer is genuinely needed otherwise login doesn't work
        headers = {"Referer": f"{Settings.base_url}/login/User/Login"}

        username, password = auth
        credentials = {
            "EM": f"{username}",  # assume email?
            "PW": f"{password}",  # password
            "ON": f"{Settings.org_number}",  # organisation number
        }

        # log in
        logger.info("Logging in")
        response = self._post(f"{Settings.base_url}/Login.ashx", headers=headers, data=credentials)

        # verify log in was successful
        props, roles = self.check_login()

        return response, props, roles

    def check_login(self, check_role_number: Optional[int] = None) -> tuple[schema.CompassProps, TYPES_ROLES_DICT]:
        """Confirms success and updates authorisation."""
        # Test 'get' for an exemplar page that needs authorisation.
        portal_url = f"{Settings.base_url}/MemberProfile.aspx?Page=ROLES&TAB"
        response = self._get(portal_url)

        # # Response body is login page for failure (~8Kb), but success is a 300 byte page.
        # if int(post_response.headers.get("content-length", 901)) > 900:
        #     raise CompassAuthenticationError("Login has failed")
        # Naive check for error, Compass redirects to an error page when something goes wrong
        # TODO what is the error page URL - what do we expect? From memory Error.aspx
        if response.url != portal_url:
            raise CompassAuthenticationError("Login has failed")

        # Create lxml html.FormElement
        form = html.fromstring(response.content).forms[0]

        # Update session dicts with new role
        compass_props = self._create_compass_props(form)  # Updates MRN property etc.
        roles_dict = dict(self._roles_iterator(form))

        member_number = compass_props.master.user.cn
        role_number = compass_props.master.user.mrn
        session_id = compass_props.master.sys.session_id

        # Set auth headers for new role
        self.update_auth_headers(member_number, role_number, session_id)

        # Update current role properties
        current_role = roles_dict[role_number]
        logger.debug(f"Using Role: {current_role[0]} ({current_role[1]})")

        # Verify role number against test value
        if check_role_number is not None:
            logger.debug("Confirming role has been changed")
            # Check that the role has been changed to the desired role. If not, raise exception.
            if check_role_number != role_number:
                raise CompassAuthenticationError("Role failed to update in Compass")

        return compass_props, roles_dict

    def update_auth_headers(self, membership_number: int, role_number: int, session_id: str) -> None:
        auth_headers = {
            "Authorization": f"{membership_number}~{role_number}",
            "SID": session_id,  # Session ID
        }
        self._update_headers(auth_headers)

    @staticmethod
    def _create_compass_props(form_tree: html.FormElement) -> schema.CompassProps:
        """Create Compass info dict from FormElement."""
        compass_props = {}
        compass_vars = form_tree.fields["ctl00$_POST_CTRL"]
        for pair in compass_vars.split("~"):
            key, value = pair.split("#", 1)
            cd_tmp = compass_props
            levels = key.split(".")
            for level in levels[:-1]:
                cd_tmp = cd_tmp.setdefault(level, {})
            cd_tmp[levels[-1]] = value  # int or str

        if "Sys" in compass_props.get("Master", {}):
            cp_m_s = compass_props["Master"]["Sys"]
            if "WebPath" in cp_m_s:
                cp_m_s["WebPath"] = urllib.parse.unquote(cp_m_s["WebPath"])
            if "HardTime" in cp_m_s:
                cp_m_s["HardTime"] = datetime.time.fromisoformat(cp_m_s["HardTime"].replace(".", ":"))

        return schema.CompassProps(**compass_props)

    @staticmethod
    def _roles_iterator(form_tree: html.FormElement) -> Iterator[tuple[int, TYPES_ROLE]]:
        """Generate role number to role name mapping."""
        roles_rows = form_tree.xpath("//tbody/tr")  # get roles from compass page (list of table rows (tr))
        for row in roles_rows:
            if "Full" in row[5].text_content():  # TODO do prov roles show up in selector???
                yield int(row.get("data-pk")), (row[0].text_content().strip(), row[2].text_content().strip())
